"""
 Copyright (C) 2015 Maruf Maniruzzaman
 Website: http://cosmosframework.com
 Author: Maruf Maniruzzaman
 License :: OSI Approved :: MIT License
"""

import jwt, Crypto.PublicKey.RSA as RSA, datetime
import uuid


class ValidationError(ValueError):
    pass

"""
  FROM RFC 6749: http://tools.ietf.org/html/rfc6749


  +--------+                                           +---------------+
  |        |--(A)------- Authorization Grant --------->|               |
  |        |                                           |               |
  |        |<-(B)----------- Access Token -------------|               |
  |        |               & Refresh Token             |               |
  |        |                                           |               |
  |        |                            +----------+   |               |
  |        |--(C)---- Access Token ---->|          |   |               |
  |        |                            |          |   |               |
  |        |<-(D)- Protected Resource --| Resource |   | Authorization |
  | Client |                            |  Server  |   |     Server    |
  |        |--(E)---- Access Token ---->|          |   |               |
  |        |                            |          |   |               |
  |        |<-(F)- Invalid Token Error -|          |   |               |
  |        |                            +----------+   |               |
  |        |                                           |               |
  |        |--(G)----------- Refresh Token ----------->|               |
  |        |                                           |               |
  |        |<-(H)----------- Access Token -------------|               |
  +--------+           & Optional Refresh Token        +---------------+

               Figure 2: Refreshing an Expired Access Token

"""

AUTH_CLIENT_APPLICATION_OBJECT_NAME = "cosmos.auth.applications"


def authorize(user, response_type, client_id, redirect_uri, **kwargs):
    state = kwargs.get("state", None)
    resource = kwargs.get("resource", "id")

    session_state = str(uuid.uuid4())
    code = {"resource": resource, }
    result = {"code": code, "session_state": session_state}

    if state:
        result["state"] = state

    return result


def get_token(**kwargs):
    service_private_pem = kwargs.get("service_private_pem")
    if not service_private_pem:
        raise ValueError("service_private_pem is not defined")

    aud = kwargs.get("aud", None)
    exp = kwargs.get("exp", None)
    family_name = kwargs.get("family_name", None)
    given_name = kwargs.get("given_name", None)

    token_payload = {
        "aud": aud,
        "exp": exp,
        "family_name": family_name,
        "given_name": given_name,
        "ver": "1.0"
    }

    iss = kwargs.get("iss", None)
    if iss:
        token_payload["iss"] = iss

    nbf = kwargs.get("nbf", None)
    if nbf:
        token_payload["nbf"] = nbf

    oid = kwargs.get("oid", None)
    if oid:
        token_payload["oid"] = oid

    sub = kwargs.get("sub", None)
    if sub:
        token_payload["sub"] = sub

    tid = kwargs.get("tid", None)
    if tid:
        token_payload["tid"] = tid

    unique_name = kwargs.get("unique_name", None)
    if unique_name:
        token_payload["unique_name"] = unique_name

    upn = kwargs.get("upn", None)
    if upn:
        token_payload["upn"] = upn

    priv_key = RSA.importKey(service_private_pem)
    token = jwt.generate_jwt(token_payload, priv_key, 'RS256', exp)

    return token


def verify_token(token, public_key_pem, alg_list):
    try:
        header, claims = jwt.verify_jwt(token, public_key_pem, alg_list)
        return header, claims
    except Exception as ex:
        raise ValidationError(ex)





"""
 Copyright (C) 2016 Maruf Maniruzzaman
 Website: http://cosmosframework.com
 Author: Maruf Maniruzzaman
 License :: OSI Approved :: MIT License
"""

import boto.sqs
import logging
from boto.sqs.message import Message


# WARNING: This is a blocking client
class AmazonSQSClient:
    def __init__(self, **kwargs):
        logging.debug("AmazonSQS Client initialization complete.")
        self.aws_access_key_id = kwargs.get("aws_access_key_id")
        self.aws_secret_access_key = kwargs.get("aws_secret_access_key")
        self.queue_region = kwargs.get("queue_region")
        self.queue_name = kwargs.get("queue_name")
        self.connection = None
        self.queue = None

    def connect(self):
        logging.debug("Connecting to AmazonSQS")
        self.connection = boto.sqs.connect_to_region(self.queue_region,
                                                     aws_access_key_id= self.aws_access_key_id,
                                                     aws_secret_access_key=self.aws_secret_access_key)

        self.queue = self.connection.create_queue(self.queue_name)

    def send_message(self, message):
        logging.debug("Sending message {}".format(message))
        m = Message()
        m.set_body(message)
        self.queue.write(m)



"""
TODO: Make async using direct http calls

http://sqs.eu-west-1.amazonaws.com/
    ?Action=CreateQueue
    &DefaultVisibilityTimeout=40
    &QueueName=testQueue
    &Version=2012-11-05
    &AUTHPARAMS

For more information:
http://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/MakingRequests_MakingQueryRequestsArticle.html
http://docs.aws.amazon.com/general/latest/gr/sigv4-signed-request-examples.html
"""


"""
 Copyright (C) 2014 Maruf Maniruzzaman
 Website: http://cosmosframework.com
 Author: Maruf Maniruzzaman
 License :: OSI Approved :: MIT License
"""

from builtins import bytes

from tornado.httpclient import AsyncHTTPClient
from tornado.template import Template
import logging
import cosmos
from cosmos.rbac.object import *

import tornado.web
from tornado import gen
from cosmos.service.requesthandler import *
import json
from cosmos.service.constants import *
import tornado.auth
from tornado.concurrent import return_future

try:
    import urlparse  # py2
except ImportError:
    import urllib.parse as urlparse  # py3

try:
    import urllib.parse as urllib_parse  # py3
except ImportError:
    import urllib as urllib_parse  # py2

USER_STATE_COOKIE_NAME = "user_state"

LOGIN_PAGE_TEMPLATE = """
<!DOCTYPE html>
<head>
    <meta charset="utf-8">
    <meta http-equiv="X-UA-Compatible" content="IE=edge">
    <title>Login</title>
    <meta name="description" content="">
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <link rel="stylesheet" href="http://ajax.googleapis.com/ajax/libs/angular_material/1.0.0/angular-material.min.css">
    <style>
        html,
        body {
            height: 100%;
        }
        html {
            display: table;
            margin: auto;
        }
        body {
            display: table-cell;
            vertical-align: middle;
        }
    </style>
</head>
<body ng-app="loginApp" ng-cloak>
<form action="." method="POST">
    <md-content class="md-padding" layout-xs="column" layout="row">
    <div layout="column">
      <md-card>
        <md-card-title>
          <md-card-title-text>
            <span class="md-headline">{{title}}</span>
          </md-card-title-text>
          <md-card-title-media>
          </md-card-title-media>
        </md-card-title>
        <md-card-content>
            {% if message %}
            <span>{{message}}</span>
            {% end %}
            <input type="hidden" name="next" value="{{ next }}" />
            <md-input-container class="md-block">
                    <label>Username</label>
                    <input required id="username" type="text" name="username" />
            </md-input-container>
            <md-input-container class="md-block">
                    <label>Password</label>
                    <input required type="password" id="password" name="password" />
            </md-input-container>
            {% if repeat_password_field %}
            <md-input-container class="md-block">
                    <label>Password again</label>
                    <input required type="password" id="password_re" name="password_re" />
            </md-input-container>
            {% end %}
        </md-card-content>
        <md-card-actions layout="row" layout-align="end center">
          <input id="loginbtn" type="submit" value="{{ btnText }}" class="md-button md-raised md-primary"></input>
        </md-card-actions>
      </md-card>
    <md-content>
    </form>
    <script src="http://ajax.googleapis.com/ajax/libs/angularjs/1.4.8/angular.min.js"></script>
    <script src="http://ajax.googleapis.com/ajax/libs/angularjs/1.4.8/angular-animate.min.js"></script>
    <script src="http://ajax.googleapis.com/ajax/libs/angularjs/1.4.8/angular-aria.min.js"></script>
    <script src="http://ajax.googleapis.com/ajax/libs/angularjs/1.4.8/angular-messages.min.js"></script>

    <!-- Angular Material Library -->
    <script src="http://ajax.googleapis.com/ajax/libs/angular_material/1.0.0/angular-material.min.js"></script>

    <!-- Your application bootstrap  -->
    <script type="text/javascript">
        angular.module('loginApp', ['ngMaterial']);
    </script>
</body>
</html>
"""


hmac_key = None

class LogoutHandler(RequestHandler):
    @gen.coroutine
    def get(self):
        self.logout_current_user()
        self.redirect('/')


def get_jwt_payload_json(id_token):
    jwt_parts = str.split(str(id_token), '.')
    pad = 4 - len(jwt_parts[1]) % 4
    pads=["", "=", "==", "===", ""]
    id_values = base64.b64decode(jwt_parts[1]+pads[pad])

    return json.loads(id_values.decode("utf-8"))


class CosmosAuthHandler(RequestHandler):
    @gen.coroutine
    def _authenticate_user(self, user_identity):
        self.current_identity = user_identity

        identity_type = user_identity.get("identity_type", None)

        #TODO: define methods and override in subclasses so we do not require these ifs and in functions below
        if identity_type == IDENTITY_TYPE_AUTHP_OAUTH2:
            iss = user_identity.get("iss")
            sub = user_identity.get("sub")
            query = {"identity_type": IDENTITY_TYPE_AUTHP_OAUTH2, "sub": sub, "iss": iss}
            columns = ["identity_type", "iss", "sub", "user_id"]
        elif identity_type == IDENTITY_TYPE_FB_GRAPH:
            fb_userid = user_identity["id"]
            query={"identity_type": IDENTITY_TYPE_FB_GRAPH, "id": fb_userid}
            columns = ["identity_type", "id", "user_id"]
        elif identity_type == IDENTITY_TYPE_GOOGLE_OAUTH2:
            id_token = user_identity["id_token"]
            id_values = get_jwt_payload_json(id_token)
            sub = id_values.get("sub")
            query = {"sub": sub, "identity_type": IDENTITY_TYPE_GOOGLE_OAUTH2}
            columns = ["identity_type", "sub", "user_id"]
        elif identity_type == IDENTITY_TYPE_GITHUB_OAUTH2:
            gh_userid = user_identity["id"]
            query={"id": gh_userid, "identity_type": IDENTITY_TYPE_GITHUB_OAUTH2}
            columns = ["identity_type", "id", "user_id"]
        else:
            claimed_id = user_identity.get("claimed_id")
            query = {"claimed_id":claimed_id}
            columns = ["identity_type", "claimed_id", "user_id"]

        object_service = self.settings['object_service']

        cursor = object_service.find(SYSTEM_USER, COSMOS_USERS_IDENTITY_OBJECT_NAME, query, columns)

        if (yield cursor.fetch_next):
            stored_user_id = cursor.next_object()
            user = yield object_service.load(SYSTEM_USER, COSMOS_USERS_OBJECT_NAME, str(stored_user_id["user_id"]), [])

            if not user:
                raise tornado.web.HTTPError(500, "User not found for previously stored identity of this user")

            self.set_current_user(user)
        else:
            create_result = yield self._create_user(user_identity)

            cursor = object_service.find(SYSTEM_USER, COSMOS_USERS_OBJECT_NAME, {"_id": create_result}, [])
            if (yield cursor.fetch_next):
                user = cursor.next_object()
            else:
                raise tornado.web.HTTPError(500, "Could not find created user")

            result = yield self._create_user_identity(str(user["_id"]), user_identity)

            self.set_current_user(user)

    def _create_user(self, user_identity):
        identity_type = user_identity.get("identity_type", None)

        if identity_type == IDENTITY_TYPE_FB_GRAPH:
            fb_userid = user_identity["id"]
            data = {"fb_userid": fb_userid, "roles": []}
        elif identity_type == IDENTITY_TYPE_GOOGLE_OAUTH2:
            id_token = user_identity["id_token"]
            id_values = get_jwt_payload_json(id_token)
            email = id_values.get("email")
            data = {"email": email, "roles": []}
        else:
            email = user_identity.get("email")
            data = {"email": email, "roles": []}

        object_service = self.settings['object_service']
        return object_service.save(SYSTEM_USER, COSMOS_USERS_OBJECT_NAME, data)

    def _create_user_identity(self, user_id, identity):
        data = identity
        data["user_id"] = user_id

        identity_type = identity.get("identity_type", None)

        if identity_type == IDENTITY_TYPE_GOOGLE_OAUTH2:
            id_token = identity["id_token"]
            id_values = get_jwt_payload_json(id_token)
            data.update(id_values)

        object_service = self.settings['object_service']
        return object_service.save(SYSTEM_USER, COSMOS_USERS_IDENTITY_OBJECT_NAME, data)


class OpenidLoginHandler(CosmosAuthHandler, tornado.auth.OpenIdMixin):
    @return_future
    def authorize_redirect(self, oauth_scope, callback_uri=None,
                           ax_attrs=["name", "email", "language", "username"],
                           callback=None):

        callback_uri = callback_uri or self.request.uri
        args = self._openid_args(callback_uri, ax_attrs=ax_attrs,
                                 oauth_scope=oauth_scope)
        self.redirect(self._OPENID_ENDPOINT + "?" + urllib_parse.urlencode(args))
        callback()

    @tornado.web.asynchronous
    @gen.coroutine
    def get(self):
        #TODO: Evaluate security risk of accepting openid.op_endpoint value of remote site response
        self._OPENID_ENDPOINT = self.get_argument("openid.op_endpoint", None)
        if not self._OPENID_ENDPOINT:
            raise tornado.web.HTTPError(400, "Endpoint not found")

        if self.get_argument("openid.mode", None):
            user = yield self.get_authenticated_user()
            if not user:
                raise tornado.web.HTTPError(500, "Openid auth failed")

            assert isinstance(user, dict)

            user['identity_type'] = IDENTITY_TYPE_OPEN_ID

            yield self._authenticate_user(user)
            self.redirect(self.get_argument("next", '/'))
        else:
            callback_uri = self.request.uri
            url = urlparse.urljoin(self.request.full_url(), callback_uri)
            urls = url  # url.replace('http','https')
            yield self.authenticate_redirect(callback_uri=urls)


class GoogleOAuth2LoginHandler(CosmosAuthHandler, tornado.auth.GoogleOAuth2Mixin):
    @tornado.web.asynchronous
    @gen.coroutine
    def get(self):
        urls = self.settings[self._OAUTH_SETTINGS_KEY]['redirect_uri']

        if self.get_argument('code', False):
            # The url must match what was used during authorize phase.
            # It can not accept if there is any mismatch including parameters
            user = yield self.get_authenticated_user(redirect_uri=urls, code=self.get_argument('code'))

            if not user:
                raise tornado.web.HTTPError(500, "Google auth failed")

            user["identity_type"] = IDENTITY_TYPE_GOOGLE_OAUTH2
            yield self._authenticate_user(user)
            self.redirect(self.get_argument("next", '/'))
        else:
            yield self.authorize_redirect(
                redirect_uri=urls,
                client_id=self.settings[self._OAUTH_SETTINGS_KEY]['key'],
                client_secret=self.settings[self._OAUTH_SETTINGS_KEY]['secret'],
                scope=['profile', 'email'],
                response_type='code',
                extra_params={'approval_prompt': 'auto'})


class GithubOAuth2LoginHandler(CosmosAuthHandler, tornado.auth.OAuth2Mixin):
    _OAUTH_AUTHORIZE_URL = "https://github.com/login/oauth/authorize"
    _OAUTH_ACCESS_TOKEN_URL = "https://github.com/login/oauth/access_token"
    _OAUTH_SETTINGS_KEY = 'github_oauth'

    @gen.coroutine
    def get_authenticated_user(self, redirect_uri, code, callback=None):
        body = urllib_parse.urlencode({
            "redirect_uri": redirect_uri,
            "code": code,
            "client_id": self.settings[self._OAUTH_SETTINGS_KEY]['client_id'],
            "client_secret": self.settings[self._OAUTH_SETTINGS_KEY]['secret'],
            "grant_type": "authorization_code",
        })
        http = tornado.httpclient.AsyncHTTPClient()

        response = yield http.fetch(self._OAUTH_ACCESS_TOKEN_URL,
                   method="POST", headers={'Content-Type': 'application/x-www-form-urlencoded', 'Accept': 'application/json'}, body=body)

        if response.error:
            tornado.AuthError('Github auth error: %s' % str(response))

        user_token = escape.json_decode(response.body)
        url = "https://api.github.com/user?access_token="+user_token.get("access_token")
        response = yield http.fetch(url,
                   method="GET", headers={"User-Agent": "Cosmos", 'Accept': 'application/json'})

        result = escape.json_decode(response.body)
        result.update(user_token)

        raise gen.Return(result)

    @tornado.web.asynchronous
    @gen.coroutine
    def get(self):
        urls = self.settings[self._OAUTH_SETTINGS_KEY]['redirect_uri']

        if self.get_argument('code', False):
            # The url must match what was used during authorize phase.
            # It can not accept if there is any mismatch including parameters
            user = yield self.get_authenticated_user(redirect_uri=urls, code=self.get_argument('code'))

            if not user:
                raise tornado.web.HTTPError(500, "Github auth failed")

            user["identity_type"] = IDENTITY_TYPE_GITHUB_OAUTH2
            yield self._authenticate_user(user)
            self.redirect(self.get_argument("next", '/'))
        else:
            yield self.authorize_redirect(
                redirect_uri=urls,
                client_id=self.settings[self._OAUTH_SETTINGS_KEY]['client_id'],
                client_secret=self.settings[self._OAUTH_SETTINGS_KEY]['secret'],
                scope=['user'],
                response_type='code',
                extra_params={'approval_prompt': 'auto'})


def add_params(input_url, params):
    url_parts = list(urlparse.urlparse(input_url))
    query = dict(urlparse.parse_qsl(url_parts[4]))
    query.update(params)
    url_parts[4] = urllib_parse.urlencode(query)
    return urlparse.urlunparse(url_parts)


class FacebookGraphLoginHandler(CosmosAuthHandler, tornado.auth.FacebookGraphMixin):
    @tornado.web.asynchronous
    @gen.coroutine
    def get(self):
        default_next = self.settings.get("default_login_next_uri", "/")
        nextpg = self.get_argument("next", default_next)
        params = {'next': nextpg}
        callback_uri = self.request.uri

        redirect_uri_base = self.settings.get("facebook_redirect_uri", self.request.full_url())

        url = urlparse.urljoin(redirect_uri_base, callback_uri)

        redirect_uri = add_params(url, params)

        redirect_uris = redirect_uri #redirect_uri.replace('http','https')

        if self.get_argument("code", False):
            user = yield self.get_authenticated_user(
                                        redirect_uri= redirect_uris,
                                        client_id=self.settings["facebook_api_key"],
                                        client_secret=self.settings["facebook_secret"],
                                        code=self.get_argument("code")
                                        )
            if user:
                user['identity_type'] = IDENTITY_TYPE_FB_GRAPH
                yield self._authenticate_user(user)
                self.redirect(self.get_argument("next", '/'))
            else:
                raise tornado.web.HTTPError(500, "Facebook auth failed")
        else:
            yield self.authorize_redirect(redirect_uri=redirect_uri,
                              client_id= self.settings["facebook_api_key"],
                              extra_params={"scope": self.settings["facebook_scope"]})


class AuthpOAuth2LoginHandler(CosmosAuthHandler):
    @gen.coroutine
    def get(self):
        server_root = "{}://{}".format(self.request.protocol, self.request.host)
        tenant_id = self.settings.get("tenant_id", "common")
        resource = self.settings.get("resource")

        code = self.get_argument("code", default=None)
        token = self.get_argument("access_token", default=None)

        state = self.get_secure_cookie(USER_STATE_COOKIE_NAME)

        redirect_uri = self.get_redirect_url()

        if code:
            token_url = self.get_token_url()
            parts = list(urlparse.urlparse(token_url))
            query = dict(urlparse.parse_qsl(parts[4]))
            params = dict(code=code, grant_type="code", redirect_uri=redirect_uri, state=state,
                          tenant_id=tenant_id)
            query.update(params)
            parts[4] = urlencode(query)
            auth_url = urlparse.urlunparse(parts)
            self.redirect(auth_url)
        elif token:
            pub_key_url = self.get_public_key_url()
            logging.debug("Downloading public key from [{0}]".format(pub_key_url))
            http_client = AsyncHTTPClient()
            resp = yield http_client.fetch(pub_key_url)

            if not resp or not resp.code == 200 or resp.body is None:
                self.write("Could not get auth server public key")
            else:
                pub_pem = resp.body
                logging.debug("Public key: {0}".format(pub_pem))
                header, user = cosmos.auth.oauth2.verify_token(token, pub_pem, ['RS256'])

                if user:
                    user['identity_type'] = IDENTITY_TYPE_AUTHP_OAUTH2
                    yield self._authenticate_user(user)
                    self.redirect(self.get_argument("next", '/'))
                else:
                    raise tornado.web.HTTPError(500, "OAuth2 auth failed")
        else:
            state = str(uuid.uuid4())
            self.set_secure_cookie(USER_STATE_COOKIE_NAME, state)

            auth_url = self.get_authorize_url()
            parts = list(urlparse.urlparse(auth_url))
            query = dict(urlparse.parse_qsl(parts[4]))
            params = dict(response_type="code", resource=server_root, redirect_uri=redirect_uri, state=state,
                          tenant_id=tenant_id)
            query.update(params)
            parts[4] = urlencode(query)
            auth_start = urlparse.urlunparse(parts)
            self.redirect(auth_start)

    def get_redirect_url(self):
        redirect_uri = "{}://{}/login/authp/".format(self.request.protocol, self.request.host)
        return redirect_uri

    def get_service_url(self):
        oauth2_settings = self.settings['oauth2_settings']
        return oauth2_settings.get("oauth2_service_url", "https://authp.com/")

    def get_authorize_url(self):
        tenant_id = self.settings.get("tenant_id", "common")
        base_url = self.get_service_url()
        return urlparse.urljoin(base_url, "{}/oauth2/authorize/".format(tenant_id))

    def get_token_url(self):
        tenant_id = self.settings.get("tenant_id", "common")
        base_url = self.get_service_url()
        return urlparse.urljoin(base_url, "{}/oauth2/token/".format(tenant_id))

    def get_token_refresh_url(self):
        tenant_id = self.settings.get("tenant_id", "common")
        base_url = self.get_service_url()
        return urlparse.urljoin(base_url, "{}/oauth2/refresh/".format(tenant_id))

    def get_public_key_url(self):
        tenant_id = self.settings.get("tenant_id", "common")
        base_url = self.get_service_url()
        return urlparse.urljoin(base_url, "{}/auth/key/".format(tenant_id))


class BasicLoginHandler(RequestHandler):
    @gen.coroutine
    def get(self):
        next = self.get_argument("next", '/')
        self._show_login_window(next)

    def _show_login_window(self, next="/", message=None, login_template=None):
        if not login_template:
            login_template = self.settings.get('login_template', LOGIN_PAGE_TEMPLATE)
        t = Template(login_template)
        html = t.generate(next=next, title="Sign In", btnText="Login",  message=message, repeat_password_field=False)
        self.write(html)
        self.finish()

    @gen.coroutine
    def post(self):
        username = self.get_argument("username", default=None)
        password = self.get_argument("password", default=None)
        redirect_url = self.get_argument("next", default='/')

        if not username or len(username) < 1:
            body = self.request.body
            if type(body) is bytes:
                body = body.decode('utf-8')
            if body and len(body) > 32:
                data = json.loads(body)
                assert isinstance(data, dict)
                username = data.get("username")
                password = data.get("password")
                redirect_url = data.get("next", "/")

        if not username or not password:
            self._show_login_window(message="Missing required values.")
            return

        object_name = COSMOS_USERS_OBJECT_NAME
        obj_serv = self.settings['object_service']

        columns = ["username", "password", "roles"]
        query = {"username": username}

        cursor = obj_serv.find(SYSTEM_USER, object_name, query, columns)

        found = yield cursor.fetch_next
        if not found:
            raise tornado.web.HTTPError(401, "Unauthorized")
        else:
            user = cursor.next_object()
            loaded_password_hash = user.get(PASSWORD_COLUMN_NAME)
            hmac_key= self.settings["hmac_key"]
            validate_password(password, loaded_password_hash, hmac_key)
            del user["password"]
            self.set_current_user(user)
            self.redirect(redirect_url)


class ChangePasswordHandler(RequestHandler):
    @gen.coroutine
    def get(self):
        next = self.get_argument("next", '/')
        self._show_change_password_window(next)

    def _show_change_password_window(self, next="/", message=None, page_template=None):
        if not page_template:
            page_template = self.settings.get('change_password_template', LOGIN_PAGE_TEMPLATE)
        t = Template(page_template)
        html = t.generate(next=next, title="Change password", btnText="Change",  message=message, repeat_password_field=True)
        self.write(html)
        self.finish()

    @gen.coroutine
    def post(self):
        username = self.get_argument("username", default=None)
        password = self.get_argument("password", default=None)
        redirect_url = self.get_argument("next", default='/')

        if not username or len(username) < 1:
            body = self.request.body
            if type(body) is bytes:
                body = body.decode('utf-8')
            if body and len(body) > 32:
                data = json.loads(body)
                assert isinstance(data, dict)
                username = data.get("username")
                password = data.get("password")
                new_password = data.get("new_password")
                password_re = data.get("password_re")
                redirect_url = data.get("next", "/")

        if not username or not password:
            self._show_change_password_window(message="Missing required values.")
            return

        if not password_re == password:
            self._show_change_password_window(message="Password did not match.")
            return

        object_name = COSMOS_USERS_OBJECT_NAME
        obj_serv = self.settings['object_service']

        columns = ["username", "password", "roles"]
        query = {"username": username}

        cursor = obj_serv.find(SYSTEM_USER, object_name, query, columns)

        found = yield cursor.fetch_next
        if not found:
            raise tornado.web.HTTPError(401, "Unauthorized")
        else:
            user = cursor.next_object()
            loaded_password_hash = user.get(PASSWORD_COLUMN_NAME)
            hmac_key= self.settings["hmac_key"]
            validate_password(password, loaded_password_hash, hmac_key)
            del user["password"]
            self.set_current_user(user)
            self.redirect(redirect_url)


class ResetPasswordHandler(RequestHandler):
    @gen.coroutine
    def get(self):
        pass


def validate_password(password, saved_password_hash, hmac_key):
    hmac_password = get_hmac_password(password, hmac_key)
    try:
        # FIXME: WARNING possible attack: use hmac.compare_digest(a, b) instead (https://docs.python.org/2/library/hmac.html)
        if not saved_password_hash == hmac_password:
            raise tornado.web.HTTPError(401, "Unauthorized")
    except:
        raise tornado.web.HTTPError(401, "Unauthorized")

def get_hmac_password(password, hmac_key):
    hmac_hex = hmac.new(hmac_key.encode("utf-8"), password.encode("utf-8")).hexdigest()
    hmac_password = "{0}{1}".format(PASSWORD_HMAC_SIGNATURE, hmac_hex)
    return hmac_password

@gen.coroutine
def before_user_insert(object_service, object_name, data, access_type):
    assert isinstance(hmac_key, object)
    assert object_name == COSMOS_USERS_OBJECT_NAME
    assert isinstance(data, dict)
    assert access_type == AccessType.INSERT or access_type == AccessType.UPDATE

    username = data.get(USERNAME_COLUMN_NAME)

    if username and len(username) > 0:
        username = username.lower()

        data["username"] = username
        query = {"username": username}
        columns=["username"]
        cursor = object_service.find(SYSTEM_USER, COSMOS_USERS_OBJECT_NAME, query, columns)

        if(yield cursor.fetch_next):
            user = cursor.next_object()
            if user:
                raise tornado.web.HTTPError(409, "Conflict: Duplicate username")

    password = data.get(PASSWORD_COLUMN_NAME)

    if not password:
        return

    if password.find(PASSWORD_HMAC_SIGNATURE) > 0:
        return

    hmac_password = get_hmac_password(password, hmac_key)

    data[PASSWORD_COLUMN_NAME] = hmac_password



"""
 Copyright (C) 2014 Maruf Maniruzzaman
 Website: http://cosmosframework.com
 Author: Maruf Maniruzzaman
 License :: OSI Approved :: MIT License
"""

import tornado.ioloop
import tornado.web
import tornado.options
import tornado.template
import tornado.websocket
from cosmos.dataservice.objectservice import *
from cosmos.datamonitor.monitor import *
from cosmos.msgq.rabbitmq import RabbitMQClient
from cosmos.service import BootLoader
import cosmos.service.auth

def init_observers(db, object_service, observers):
    loader = BootLoader()
    loader.init_observers(object_service, observers)
    loader.load_roles(object_service)
    loader.load_role_groups(object_service)


def create_message_channel():
    mq = RabbitMQClient(queue_name="worker_q")
    mq.connect()
    return mq

def start_web_service(options):

    cosmos.service.auth.hmac_key = options.hmac_key
    object_service = ObjectService(rbac_service=RbacService(), db=options.db)

    init_observers(options.db, object_service, options.observers)

    app_settings = dict(
                db=options.db,
                login_url=options.login_url,
                hmac_key=options.hmac_key,
                cookie_secret=options.cookie_secret,
                xheaders=options.xheaders,
                template_path=options.template_path,
                debug=options.debug,
                facebook_api_key=options.facebook_api_key,
                facebook_secret=options.facebook_secret,
                facebook_scope=options.facebook_scope,
                google_oauth=options.google_oauth,
                github_oauth=options.github_oauth,
                directory_listing_allowed=options.directory_listing_allowed,
                object_service=object_service,
                source_root=options.source_root,
                facebook_redirect_uri=options.facebook_redirect_uri,
                default_login_next_uri=options.default_login_next_uri,
                tenant_id=options.tenant_id,
                oauth2_settings=options.oauth2_settings,
            )

    application = tornado.web.Application(
        options.endpoints,
        **app_settings
    )

    logging.info('Starting up server on port: {0}'.format(options.web_service_port))
    application.listen(options.web_service_port)



"""
 Copyright (C) 2016 Maruf Maniruzzaman
 Website: http://cosmosframework.com
 Author: Maruf Maniruzzaman
 License :: OSI Approved :: MIT License
"""

from azure.storage.queue import QueueService
import logging


class AzureQueueClient:
    def __init__(self, **kwargs):

        self.account_name = kwargs.get("account_name")
        self.account_key = kwargs.get("account_key")

        self.queue_name = kwargs.get("queue_name")

        self.queue_service = None
        self.queue = None

        logging.debug("AzureQueue Client initialization complete.")


    def connect(self):
        logging.debug("Connecting to AzureQueue")
        queue_service = QueueService(account_name=self.account_name, account_key=self.account_key)
        self.queue_service.create_queue(self.queue_name)

    def send_message(self, message):
        logging.debug("Sending message {}".format(message))


"""
Queue Service REST API
https://msdn.microsoft.com/en-us/library/azure/dd179363


http://<storage account>.queue.core.windows.net/<queue>

https://github.com/Azure/azure-storage-python/blob/master/azure/storage/queue/queueservice.py

"""


"""
 Copyright (C) 2014 Maruf Maniruzzaman
 Website: http://cosmosframework.com
 Author: Maruf Maniruzzaman
 License :: OSI Approved :: MIT License
"""
from argparse import ArgumentError
import hashlib

import datetime
import gridfs
import motor
import tornado.web
from bson import ObjectId
from cosmos.rbac.service import *

#TODO: May be use observers to check access - initialize in the __init__ of the rbac package

ACCESS_TYPE_ROLE = 1
ACCESS_TYPE_OWNER_ONLY = 2

class ObjectService():
    def __init__(self, *args, **kwargs):
        self.rbac_service = kwargs.get("rbac_service", RbacService())
        self.db = kwargs.get("db", None)

        if not self.db:
            raise ArgumentError("db", "Database object must be set while creating ObjectService.")

        self._preprocessors = {}
        self._processors = {}
        self._postprocessors = {}

    def check_access(self, user, object_name, properties, access, check_owner=False):
        roles = self.rbac_service.get_roles(user)

        # We must check all roles for possible access before checking owner access, so we need to loop twice.
        # Since owner access is suggested rare optimizing for role access here,
        for role in roles:
            has_access = self.rbac_service.has_access(role, object_name, properties, access)
            if has_access:
                logging.debug("ObjectService:: check _access {0} is granted to {1} as role accessible for properties {2}.".format(object_name, user, properties))
                return ACCESS_TYPE_ROLE

        for role in roles:
            if check_owner:
                has_owner_access = self.rbac_service.has_owner_access(role, object_name, properties, access)
                if has_owner_access:
                    logging.debug("ObjectService:: check _access {0} is granted to {1} as owner accessible for properties {2}.".format(object_name, user, properties))
                    return ACCESS_TYPE_OWNER_ONLY

        logging.warn("ObjectService:: check _access {0} is DENIED to {1} for properties {2}.".format(object_name, user, properties))
        raise tornado.web.HTTPError(401, "Unauthorized")

    def save(self, user, object_name, data):
        logging.debug("ObjectService::save::{0}".format(object_name))
        assert isinstance(data, dict)

        properties = self.get_properties(data)
        self.check_access(user, object_name, properties, AccessType.INSERT, True)

        self.create_access_log(user, object_name, AccessType.INSERT)

        data['createtime'] = str(datetime.datetime.now())

        if user:
            data['owner'] = str(user.get("_id"))
        else:
            data['owner'] = SYSTEM_USER

        if data.get("_id"):
            data["_id"] = ObjectId(str(data["_id"]))

        #TODO: Make sure user can not insert data for object owned by other user. With _id it may be possible.

        result = self.db[object_name].save(data)

        return result

    def insert(self, user, object_name, data):
        logging.debug("ObjectService::insert::{0}".format(object_name))
        assert isinstance(data, dict)

        properties = self.get_properties(data)
        self.check_access(user, object_name, properties, AccessType.INSERT, True)

        self.create_access_log( user, object_name, AccessType.INSERT)

        data['createtime'] = str(datetime.datetime.now())

        if user:
            data['owner'] = str(user.get("_id"))
        else:
            data['owner'] = SYSTEM_USER

        result = self.db[object_name].insert(data)

        return result

    def find(self, user, object_name, query, columns, limit=5000):
        logging.debug("ObjectService::find::{0}".format(object_name))
        #assert inspect.ismethod(callback)

        allowed_access_type = self.check_access(user, object_name, columns, AccessType.READ, True)

        if allowed_access_type == ACCESS_TYPE_OWNER_ONLY:
            if query:
                assert isinstance(query, dict)
                query["owner"] = str(user.get("_id"))
            else:
                query = {"owner": str(user.get("_id"))}

        self.create_access_log(user, object_name, AccessType.READ)

        if len(columns) > 0:
            columns_dict = {column:1 for column in columns}
        else:
            columns_dict = None

        result = self.db[object_name].find(query, columns_dict).limit(limit)

        return result

    def text_search(self, user, object_name, query, columns, limit=5000):
        logging.debug("ObjectService::text_search::{0}".format(object_name))

        allowed_access_type = self.check_access(user, object_name, columns, AccessType.SEARCH, True)

        if allowed_access_type == ACCESS_TYPE_OWNER_ONLY:
            if query:
                assert isinstance(query, dict)
                query["owner"] = str(user.get("_id"))
            else:
                query = {"owner": str(user.get("_id"))}

        self.create_access_log(user, object_name, AccessType.SEARCH)

        if len(columns) > 0:
            columns_dict = {column:1 for column in columns}
        else:
            columns_dict = None

        result = self.db[object_name].find(query, columns_dict).limit(limit)

        return result

    def load(self, user, object_name, id, columns):
        logging.debug("ObjectService::load::{0}".format(object_name))

        allowed_access_type = self.check_access(user, object_name, columns, AccessType.READ, True)

        if len(columns) > 0:
            columns_dict = {column:1 for column in columns}
        else:
            columns_dict = None

        query = {'_id': ObjectId(id)}

        if allowed_access_type == ACCESS_TYPE_OWNER_ONLY:
            query["owner"] = str(user.get("_id"))

        self.create_access_log(user, object_name, AccessType.READ)

        result = self.db[object_name].find_one(query, columns_dict)
        return result

    # TODO: SECURITY_ISSUE: escape $ sign in values
    # http://docs.mongodb.org/manual/faq/developers/#dollar-sign-operator-escaping
    def update(self, user, object_name, id, data):
        logging.debug("ObjectService::update::{0}".format(object_name))
        logging.debug("Data: {0}".format(data))
        assert len(id) > 0
        assert isinstance(data, dict)

        properties = self.get_properties(data)

        allowed_access_type = self.check_access(user, object_name, properties, AccessType.UPDATE, True)

        query = {'_id': ObjectId(id)}
        if allowed_access_type == ACCESS_TYPE_OWNER_ONLY:
            query["owner"] = str(user.get("_id"))

        self.create_access_log(user, object_name, AccessType.UPDATE)

        result = self.db[object_name].update(query, {'$set': data})

        return result

    def delete(self, user, object_name, id):
        logging.debug("ObjectService::delete::{0}".format(object_name))
        assert len(id) > 0

        allowed_access_type = self.check_access(user, object_name, [], AccessType.DELETE, True)

        query = {'_id': ObjectId(id)}
        if allowed_access_type == ACCESS_TYPE_OWNER_ONLY:
            query["owner"] = str(user.get("_id"))

        self.create_access_log(user, object_name, AccessType.DELETE)

        result = self.db[object_name].remove(query)

        return result

    def save_file(self, user, collection_name, file_object, file_id=None):
        logging.debug("ObjectService::save_file::{0}".format(collection_name))

        properties = ['body', 'content_type', 'filename', 'collection_name', 'createtime', 'owner']

        if file_id:
            self.check_access(user, collection_name, properties, AccessType.UPDATE, True)
        else:
            self.check_access(user, collection_name, properties, AccessType.INSERT, True)

        if file_id:
            self.create_access_log(user, collection_name, AccessType.UPDATE)
        else:
            self.create_access_log(user, collection_name, AccessType.INSERT)

        return save_file_in_gridfs(self.db, user, collection_name, file_object, properties, file_id)

    def load_file(self, user, collection_name, file_id, ignore_col_name=False):
        logging.debug("ObjectService::load_file::{0}".format(collection_name))

        properties = ['body', 'content_type']

        if not ignore_col_name:
            allowed_access_type = self.check_access(user, collection_name, properties, AccessType.READ, True)

            if allowed_access_type == ACCESS_TYPE_OWNER_ONLY:
                return read_gridfs_owned_file(user, self.db, collection_name, file_id, properties)

            self.create_access_log(user, collection_name, AccessType.READ)
        else:
            self.create_access_log(user, "gridfile:"+file_id, AccessType.READ)

        return read_gridfs_file(self.db, collection_name, file_id, properties)

    @gen.coroutine
    def list_file(self, user, collection_name):
        logging.debug("ObjectService::save_file::{0}".format(collection_name))
        properties = ['body', 'content_type', 'filename', 'collection_name', 'createtime', 'owner', 'md5', 'length']

        allowed_access_type = self.check_access(user, collection_name, properties, AccessType.READ, True)

        query = {'collection_name': collection_name}
        if allowed_access_type == ACCESS_TYPE_OWNER_ONLY:
            query["owner"] = str(user.get("_id"))

        file_list = []

        fs = motor.MotorGridFS(self.db)
        cursor = fs.find(query, timeout=False)
        while (yield cursor.fetch_next):
            grid_out = cursor.next_object()
            file_list.append({"file_id": grid_out._id, "filename":grid_out.filename,
                              'content_type':grid_out.content_type, 'owner':grid_out.owner, 'md5':grid_out.md5,
                              'createtime':grid_out.createtime, "length": grid_out.length,
                              "collection_name":collection_name})
            grid_out.close()

        raise gen.Return(file_list)

    def delete_file(self, user, collection_name, file_id):
        logging.debug("ObjectService::delete_file::{0}".format(collection_name))
        assert len(file_id) > 0

        allowed_access_type = self.check_access(user, collection_name, [], AccessType.DELETE, True)

        self.create_access_log(user, collection_name, AccessType.DELETE)

        if allowed_access_type == ACCESS_TYPE_OWNER_ONLY:
            return delete_gridfs_owned_file(user, self.db, file_id)

        return delete_gridfs_file(self.db, file_id)

    def get_properties(self, data, namespace=None):
        properties = list(data.keys())
        child_props = []

        for prop in properties:
            prop_data = data.get(prop, None)

            if prop_data and isinstance(prop_data, dict):
                child_namespace = (namespace + "." + prop) if namespace else prop
                child_props = child_props + self.get_properties(prop_data, child_namespace)

        if child_props:
            properties = properties + child_props

        properties = [p if not namespace else namespace + "." + p for p in properties]

        return properties

    def create_access_log(self, user, module, function):
        username = None
        access_log = {"username":username, 'module': module, 'function': function}
        access_log['createtime'] = str(datetime.datetime.now())
        #db.audit.access_log.insert(access_log, callback=None)

    def add_operation_preprocessor(self, preprocessor, object_name, access_types):
        assert isinstance(access_types, collections.Iterable)

        assert len(object_name) > 0

        object_preprocessor =self._preprocessors.get(object_name)
        if not object_preprocessor:
            object_preprocessor = {}
            self._preprocessors[object_name] = object_preprocessor

        for access_type in access_types:
            object_acctyp_preprocessor = object_preprocessor.get(access_type)
            if not object_acctyp_preprocessor:
                object_acctyp_preprocessor = []
                object_preprocessor[access_type] = object_acctyp_preprocessor

            if not preprocessor in object_acctyp_preprocessor:
                object_acctyp_preprocessor.append(preprocessor)

    def add_operation_processor(self, processor, object_name, access_types):
        assert isinstance(access_types, collections.Iterable)

        assert len(object_name) > 0

        object_processor =self._processors.get(object_name)
        if not object_processor:
            object_processor = {}
            self._processors[object_name] = object_processor

        for access_type in access_types:
            object_acctyp_processor = object_processor.get(access_type)
            if not object_acctyp_processor:
                object_acctyp_processor = []
                object_processor[access_type] = object_acctyp_processor

            if not processor in object_acctyp_processor:
                object_acctyp_processor.append(processor)

    def add_operation_postprocessor(self, postprocessor, object_name, access_types):
        #assert inspect.ismethod(preprocessor)
        assert isinstance(access_types, collections.Iterable)

        assert len(object_name) > 0

        object_postprocessor =self._postprocessors.get(object_name)
        if not object_postprocessor:
            object_postprocessor = {}
            self._postprocessors[object_name] = object_postprocessor

        for access_type in access_types:
            object_acctype_postprocessor = object_postprocessor.get(access_type)
            if not object_acctype_postprocessor:
                object_acctype_postprocessor = []
                object_postprocessor[access_type] = object_acctype_postprocessor

            if not postprocessor in object_acctype_postprocessor:
                object_acctype_postprocessor.append(postprocessor)

    def get_operation_preprocessor(self, object_name, access_type):
        assert len(object_name) > 0
        object_preprocessor = self._preprocessors.get(object_name)

        if not object_preprocessor:
            return []

        preprocessor_list = object_preprocessor.get(access_type)
        if not preprocessor_list:
            return []

        return preprocessor_list

    def get_operation_processor(self, object_name, access_type):
        assert len(object_name) > 0
        object_processor = self._processors.get(object_name)

        if not object_processor:
            return []

        processor_list = object_processor.get(access_type)
        if not processor_list:
            return []

        return processor_list

    def get_operation_postprocessor(self, object_name, access_type):
        assert len(object_name) > 0
        object_postprocessor = self._postprocessors.get(object_name)

        if not object_postprocessor:
            return []

        post_processor_list = object_postprocessor.get(access_type)
        if not post_processor_list:
            return []

        return post_processor_list

@gen.coroutine
def save_file_in_gridfs(db, user, collection_name, file_object, properties, file_id=None):
    fs = motor.MotorGridFS(db)
    if file_id:
        #TODO: Verify: As of today mongodb does not allow replacing file- so delete and create
        #TODO: non atomic operation
        yield delete_gridfs_file(db, file_id)
        oid = ObjectId(str(file_id))
        gridin = yield fs.new_file(_id=oid)
    else:
        gridin = yield fs.new_file()

    file_body = file_object.get("body")
    length = len(file_body)
    md5_dig = hashlib.md5(file_body).hexdigest()
    result = yield gridin.write(file_body)

    # TODO: we can write another chunk- as many times we want-
    # When support for streaming file comes in mainstram tornado branch
    # we should use that

    if 'content_type' in properties:
        yield gridin.set('content_type', file_object.get("content_type"))

    yield gridin.set('filename', file_object.get("filename"))
    yield gridin.set('collection_name', collection_name)

    current_time = str(datetime.datetime.now())
    yield gridin.set('createtime', current_time)
    yield gridin.set('owner',str(user.get("_id")))

    yield gridin.close()

    file_id = gridin._id

    filename = gridin.filename
    result = {"md5":md5_dig, "file_id":str(file_id),"length": length, 'owner': str(user.get("_id")),
                        "collection_name":collection_name, "filename": filename,'createtime': current_time }

    raise gen.Return(result)

@gen.coroutine
def read_gridfs_owned_file(user, db, collection_name, file_id, properties):
    owner_id = user.get("_id")

    if len(owner_id) < 1:
        raise tornado.web.HTTPError(401, "Unauthorized")

    fs = motor.MotorGridFS(db)
    try:
        gridout = yield fs.get(ObjectId(file_id))

        if not gridout:
            raise tornado.web.HTTPError(404, "File not found")

        owner = gridout.owner

        if owner != owner_id:
            raise tornado.web.HTTPError(401, "Unauthorized")

        content_type = gridout.content_type
        got_col_name = gridout.collection_name

        if got_col_name != collection_name:
            raise tornado.web.HTTPError(404, "File not found")


        content = yield gridout.read()

        data = {"body": content }

        if "content_type" in properties:
            data["content_type"] = content_type

        raise gen.Return(data)
    except gridfs.NoFile:
        raise tornado.web.HTTPError(404, "File not found")

@gen.coroutine
def read_gridfs_file(db, collection_name, file_id, properties, ignore_col_name=False):
    fs = motor.MotorGridFS(db)
    try:
        gridout = yield fs.get(ObjectId(file_id))

        if not gridout:
            raise tornado.web.HTTPError(404, "File not found")

        content_type = gridout.content_type

        if not ignore_col_name:
            got_col_name = gridout.collection_name

            if got_col_name != collection_name:
                raise tornado.web.HTTPError(404, "File not found")

        content = yield gridout.read()

        data = {"body": content }

        if "content_type" in properties:
            data["content_type"] = content_type

        raise gen.Return(data)
    except gridfs.NoFile:
        raise tornado.web.HTTPError(404, "File not found")

@gen.coroutine
def delete_gridfs_file(db, file_id):
    fs = motor.MotorGridFS(db)
    result = yield fs.delete(ObjectId(file_id))
    raise gen.Return(result)

@gen.coroutine
def delete_gridfs_owned_file(user, db, file_id):

    owner_id = user.get("_id")

    if len(owner_id) < 1:
        raise tornado.web.HTTPError(401, "Unauthorized")

    fs = motor.MotorGridFS(db)

    gridout = yield fs.get(ObjectId(file_id))
    if not gridout:
        raise tornado.web.HTTPError(404, "File not found")

    owner = gridout.owner
    gridout.close()

    if owner == owner_id:
        result = yield fs.delete(ObjectId(file_id))
        raise gen.Return(result)
    else:
        raise tornado.web.HTTPError(401, "Unauthorized")







"""
 Copyright (C) 2016 Maruf Maniruzzaman
 Website: http://cosmosframework.com
 Author: Maruf Maniruzzaman
 License :: OSI Approved :: MIT License
"""

import pika
import logging
import time


class RabbitMQClient:
    def __init__(self, **kwargs):
        self.connection = None
        self.channel = None
        self.queue_name = kwargs.get("queue_name", "worker")
        logging.debug("RabbitMQ Client initialization complete.")

    def connect(self):
        credentials = pika.PlainCredentials("guest", "guest")
        params = pika.ConnectionParameters(credentials=credentials)

        pika.TornadoConnection(params, on_open_callback=self.on_open,
                               on_open_error_callback=self.on_open_error,
                               on_close_callback=self.on_close,
                               stop_ioloop_on_close=False)

    def on_open(self, connection):
        logging.info("RabbitMQ connection opened.")
        self.connection = connection
        connection.channel(self.on_channel_create)

    def on_open_error(self, connection, messae):
        logging.error("Error opening RabbitMQ connection. {}".format(messae))

    def on_channel_create(self, channel):
        logging.info("RabbitMQ channel created.")
        self.channel = channel

        channel.queue_declare(queue=self.queue_name, callback=self.on_queue_declare)

    def on_queue_declare(self, arg):
        logging.debug("RabbitMQ queue declared {}".format(arg))
        self.send_message("Hello world! " + str(time.time()))

    def send_message(self, message):
        logging.debug("Sending message {}".format(message))
        self.channel.basic_publish(exchange='',routing_key=self.queue_name, body=message)

    def on_close(self, connection):
        logging.info("RabbitMQ connection closed. Will reopen now.")
        self.connect()



"""
 Copyright (C) 2014 Maruf Maniruzzaman
 Website: http://cosmosframework.com
 Author: Maruf Maniruzzaman
 License :: OSI Approved :: MIT License
"""

import collections
from mongolog import MongoHandler
import tornado
import logging
import pymongo
import cosmos
import cosmos.dataservice.objectservice
from cosmos.dataservice.objectservice import ObjectService
from cosmos.rbac.object import *
import cosmos.rbac.service
import cosmos.rbac.object
from tornado import gen
import cosmos.service.auth
import cosmos.processors
import json

OBSERVER_PRE_PROCESSOR = 0
OBSERVER_POST_PROCESSOR = 1
OBSERVER_PROCESSOR = 2

class BootLoader():

    def init_observers(self, object_service, observers):
        logging.info("init_observers")
        object_service.add_operation_preprocessor(cosmos.service.auth.before_user_insert,
                                                                    COSMOS_USERS_OBJECT_NAME,
                                                                    [AccessType.INSERT, AccessType.UPDATE])

        object_service.add_operation_preprocessor(cosmos.processors.before_role_insert,
                                                                    cosmos.rbac.object.COSMOS_ROLE_OBJECT_NAME,
                                                                    [AccessType.INSERT])

        object_service.add_operation_postprocessor(after_role_insert_update_delete,
                                                        cosmos.rbac.object.COSMOS_ROLE_OBJECT_NAME,
                                                        [AccessType.INSERT, AccessType.UPDATE, AccessType.DELETE])

        object_service.add_operation_postprocessor(after_role_group_insert_update_delete,
                                                        cosmos.rbac.object.COSMOS_ROLE_GROUP_OBJECT_NAME,
                                                        [AccessType.INSERT, AccessType.UPDATE, AccessType.DELETE])

        for observer in observers:
            assert isinstance(observer, dict)
            func = observer["function"]
            object_name = observer["object_name"]
            access = observer["access"]
            observer_type = observer.get("type", OBSERVER_PRE_PROCESSOR)
            assert isinstance(access, collections.Iterable)

            logging.info("Adding observer {}, {},{},{}".format(observer_type, func, object_name, access))

            if observer_type == OBSERVER_PRE_PROCESSOR:
                object_service.add_operation_preprocessor(func, object_name, access)
            elif observer_type == OBSERVER_POST_PROCESSOR:
                object_service.add_operation_postprocessor(func, object_name, access)
            elif observer_type == OBSERVER_PROCESSOR:
                object_service.add_operation_processor(func, object_name, access)

    @gen.coroutine
    def load_roles(self, object_service):
        logging.debug("Loading all roles.")

        object_name = COSMOS_ROLE_OBJECT_NAME
        columns = []

        cursor = object_service.find(SYSTEM_USER,  object_name, None, columns)

        while(yield cursor.fetch_next):
            role = cursor.next_object()
            try:
                logging.debug(role)
                object_service.rbac_service.update_role_cache(role)
            except ValueError as ve:
                logging.exception("Role {0} could not be loaded.".format(role.get("name")))

    @gen.coroutine
    def load_role_groups(self, object_service):
        logging.debug("Loading all role groups.")

        rg_object_name = COSMOS_ROLE_GROUP_OBJECT_NAME
        columns = []

        rg_cursor = object_service.find(SYSTEM_USER, rg_object_name, None, columns)

        while(yield rg_cursor.fetch_next):
            role_group_def = rg_cursor.next_object()
            try:
                object_service.rbac_service.update_role_group_cache(role_group_def)
            except ValueError as ve:
                logging.exception("Role group {0} could not be loaded.".format(role_group_def.get("name")))

@gen.coroutine
def after_role_insert_update_delete(object_service, object_name, result, access_type):
    assert object_name == COSMOS_ROLE_OBJECT_NAME
    assert access_type == AccessType.INSERT or access_type == AccessType.UPDATE or access_type == AccessType.DELETE
    logging.info("Role has changed. Operation = [{}]. Reloading roles.".format(access_type))
    loader = BootLoader()
    object_service.rbac_service.clear_non_system_roles()
    loader.load_roles(object_service)

@gen.coroutine
def after_role_group_insert_update_delete(object_service, object_name, result, access_type):
    assert object_name == COSMOS_ROLE_GROUP_OBJECT_NAME
    assert access_type == AccessType.INSERT or access_type == AccessType.UPDATE or access_type == AccessType.DELETE
    logging.info("Role group has changed. Operation = [{}]. Reloading roles.".format(access_type))
    loader = BootLoader()
    object_service.rbac_service.clear_non_system_role_groups()
    loader.load_role_groups(object_service)


"""
 Copyright (C) 2014 Maruf Maniruzzaman
 Website: http://cosmosframework.com
 Author: Maruf Maniruzzaman
 License :: OSI Approved :: MIT License

 This service handler follows the style found in Azure (https://msdn.microsoft.com/en-us/library/azure/dn645542.aspx)
 Also for more information look at The OAuth 2.0 Authorization Framework (http://tools.ietf.org/html/rfc6749)
"""

import ast
import base64
import json

import datetime
import time
import uuid

import tornado.web
from Crypto.PublicKey import RSA
from tornado import gen
import logging

from cosmos.rbac.object import COSMOS_USERS_OBJECT_NAME
from cosmos.auth.oauth2 import get_token, AUTH_CLIENT_APPLICATION_OBJECT_NAME
from cosmos.rbac.object import SYSTEM_USER

try:
    import urlparse  # py2
    from urllib import urlencode
    import urllib as urllib_parse
except ImportError:
    import urllib.parse as urlparse  # py3
    from urllib.parse import urlencode
    import urllib.parse as urllib_parse

from cosmos.service.requesthandler import RequestHandler

OAUTH2_REQUESTS_OBJECT_NAME = "cosmos.auth.oauth2.requests"
OAUTH2_TOKENS_OBJECT_NAME = "cosmos.auth.oauth2.tokens"
OAUTH2_CODE_STATUS_OBJECT_NAME = "cosmos.auth.oauth2.codestatus"
OAUTH2_MAX_CODE_RESPONSE_LENGTH = 200

class OAuth2RequestException(Exception):
    pass

#TODO: security check for entire class

class OAuth2ServiceHandler(RequestHandler):

    @gen.coroutine
    def get(self, tenant_id, function):
        try:
            req_id = None
            auth_request = None
            serve_request = self.get_argument("serve_request", None)
            if not (serve_request):
                auth_request = self._collect_auth_request_parameters(tenant_id, function)

                req_id = yield self.insert_auth_request(auth_request)

                if not req_id:
                    logging.critical("Could not save OAuth2 request to database.")
                    raise tornado.web.HTTPError(500, "Server error")

                logging.debug("Request saved in db {0}".format(str(req_id)))
            else:
                req_id = self.get_argument("serve_request")
                auth_request = yield self.load_auth_request(req_id)
                if not auth_request:
                    self.write("Could not find auth request. Correlation id {}".format(req_id))
                    self.finish()
                    return

            user = self.get_current_user()

            if not user:
                url = self.request.uri
                parts = list(urlparse.urlparse(url))
                query = {"serve_request": str(req_id)}
                parts[4] = urlencode(query)

                redirect_url = urlparse.urlunparse(parts)

                self.initiate_login(redirect_url)

            if function == "authorize":
                yield self._do_authorize(user, auth_request)
                return
            elif function == "token":
                yield self._do_token(user, auth_request)
                return
            else:
                raise tornado.web.HTTPError(404, "Not found")
        except OAuth2RequestException as re:
            self.clear()
            self.set_status(400)
            self.finish("<html><body>400 Bad Request<br/>{}</body></html>".format(str(re)))

    @gen.coroutine
    def _do_authorize(self, user, params):
        response_type = params.get("response_type", None)
        if not "code" == response_type:
            raise OAuth2RequestException("Response type must be code")

        client_id = params.get("client_id", None)
        redirect_uri = params.get("redirect_uri", None)
        state = params.get("state", None)
        resource = params.get("resource", None)

        trusted = yield self.is_trusted_redirect_url(redirect_uri)

        if not trusted:
            raise OAuth2RequestException("Redirect URL is not trusted. URL = {}".format(redirect_uri))

        code_attributes = ["user_id", "client_id", "resource", "iat", "code_status_id"]

        current_utc_time = self.get_current_utc_time()

        response = {}
        response.update({"user_id": user["_id"]})
        response["client_id"] = client_id
        response["resource"] = resource
        response["iat"] = current_utc_time

        code_status_id = yield self.insert_code_status(response)
        if not code_status_id:
            logging.critical("Could not save [OAuth2 Code] to database.")
            raise tornado.web.HTTPError(500, "Server error")

        response["code_status_id"] = str(code_status_id)

        code_response = { k: response.get(k) for k in code_attributes }

        str_response = json.dumps(code_response)
        bytes_resp = str_response.encode()

        if(len(bytes_resp) > OAUTH2_MAX_CODE_RESPONSE_LENGTH):
            raise tornado.web.HTTPError(414, "Request-URI Too Long")

        oauth2_public_key_handler = self.get_public_key_handler()
        enc_response = oauth2_public_key_handler.encrypt(bytes_resp, 32)

        b64_enc_response_bytes = base64.urlsafe_b64encode(enc_response[0])
        b64_enc_response = b64_enc_response_bytes.decode()

        url = redirect_uri
        session_state = str(user.get("session_id", str(uuid.uuid4())))

        params = {response_type: b64_enc_response, "session_state": session_state}
        if state:
            params["state"] = state

        parts = list(urlparse.urlparse(url))
        query = dict(urlparse.parse_qsl(parts[4]))
        query.update(params)
        parts[4] = urlencode(query)

        redirect_url = urlparse.urlunparse(parts)

        self.redirect(redirect_url)

    @gen.coroutine
    def _do_token(self, requesting_user, params):
        code = params.get("code", None)
        grant_type = params.get("grant_type", None)
        redirect_uri = params.get("redirect_uri", None)
        resource = params.get("resource", None)
        state = params.get("state", None)
        client_id = params.get("client_id", None)
        client_secret = params.get("client_secret", None)
        refresh_token = params.get("refresh_token", None)
        tid = params.get("tenant_id", None)
        token_type = "bearer"

        # if grant_type == code then code must not be empty
        # if grant_type == refresh_token then refresh_token must not be empty
        if grant_type == "code":
            if not code:
                raise OAuth2RequestException("Code must not be empty when grant_type is code")
        elif grant_type == "refresh_token":
            if not refresh_token:
                raise OAuth2RequestException("Code must not be empty when grant_type is refresh_token")
        else:
            raise OAuth2RequestException("Value of grant_type must be either code or refresh_token")

        request_host = params.get("request_host")

        oauth2_settings = self.settings['oauth2_settings']
        oauth2_token_issuer = oauth2_settings.get("oauth2_token_issuer", request_host)
        oauth2_token_expiry_seconds = oauth2_settings.get("oauth2_token_expiry_seconds")

        exp = datetime.timedelta(seconds=oauth2_token_expiry_seconds)

        enc_code = base64.urlsafe_b64decode(code.encode())

        oauth2_private_key_handler = self.get_private_key_handler()

        code_json_bytes = oauth2_private_key_handler.decrypt(enc_code)
        code_json = code_json_bytes.decode()
        code_dict = json.loads(code_json)

        user_id = code_dict.get("user_id")
        code_iat = code_dict.get("iat")

        if not user_id or not code_iat:
            raise OAuth2RequestException("Invalid code")

        user = yield self.load_user(user_id)
        if not user:
            raise OAuth2RequestException("Invalid code")

        #TODO: allow use of secret - otherwise its not secure
        if client_id and client_secret:
            app = yield self.load_client_application(client_id)
            if not app:
                raise OAuth2RequestException("Invalid application (client_id/client_secret)")

        if(not user_id or user_id != requesting_user.get("_id")):
            raise OAuth2RequestException("Invalid request. Requesting user not found or is different than logged in user.")

        #Make sure the code was not used earlier - otherwise its not secure
        code_status_id = code_dict.get("code_status_id")
        if not code_status_id:
            raise OAuth2RequestException("Invalid code")

        code_status = yield self.load_code_status(code_status_id)
        if not code_status:
            raise OAuth2RequestException("Invalid code. Possibly server error.")

        if code_status.get("used_at"):
            code_status["duplicate_attempt"] = True
            yield self.update_code_status(code_status)
            raise OAuth2RequestException("Code already used")

        current_utc_time = self.get_current_utc_time()

        code_status["used_at"] = current_utc_time
        saved = yield self.update_code_status(code_status)
        if not saved:
            logging.error("Could not save code status Id={}".format(code_status_id))

        resource = code_dict.get("resource")
        oauth2_private_key_pem = self.get_private_key_pem()

        response = get_token(aud=client_id,
                                     exp=exp,
                                     family_name=user.get("family_name"),
                                     given_name=user.get("given_name"),
                                     iat=str(current_utc_time),
                                     iss=oauth2_token_issuer,
                                     nbf=str(current_utc_time),
                                     oid=str(user.get("_id")),
                                     sub=str(user.get("_id")),
                                     tid=tid,
                                     unique_name=user.get("username"),
                                     upn=user.get("username"),
                                     service_private_pem=oauth2_private_key_pem)
        url = redirect_uri
        session_state = user.get("session_id", str(uuid.uuid4()))
        response_type = "access_token"
        params = {response_type: response, "session_state": session_state, "token_type": token_type, "resource": resource}

        if state:
            params["state"] = state

        parts = list(urlparse.urlparse(url))
        query = dict(urlparse.parse_qsl(parts[4]))
        query.update(params)
        parts[4] = urlencode(query)

        redirect_url = urlparse.urlunparse(parts)

        self.redirect(redirect_url)

    def get_current_utc_time(self):
        return int(time.time())

    def get_public_key_handler(self):
        oauth2_settings = self.settings['oauth2_settings']
        oauth2_public_key_pem = oauth2_settings.get("oauth2_public_key_pem")
        oauth2_public_key_handler = RSA.importKey(oauth2_public_key_pem)
        return oauth2_public_key_handler

    def get_private_key_handler(self):
        oauth2_settings = self.settings['oauth2_settings']
        oauth2_private_key_pem = oauth2_settings.get("oauth2_private_key_pem")
        oauth2_private_key_handler = RSA.importKey(oauth2_private_key_pem)
        return oauth2_private_key_handler

    def _collect_auth_request_parameters(self, tenant_id, function):
        params = {k: self.get_argument(k) for k in self.request.arguments}
        params["request_protocol"] = self.request.protocol
        params["request_host"] = self.request.host
        params["request_uri"] = self.request.uri
        params["tenant_id"] = tenant_id
        params["function"] = function

        return params

    @gen.coroutine
    def insert_auth_request(self, params):
        obj_serv = self.settings['object_service']
        promise = obj_serv.insert(SYSTEM_USER, OAUTH2_REQUESTS_OBJECT_NAME, params)
        req_id = yield promise
        raise gen.Return(req_id)

    @gen.coroutine
    def load_auth_request(self, req_id):
        obj_serv = self.settings['object_service']
        cursor = obj_serv.load(SYSTEM_USER, OAUTH2_REQUESTS_OBJECT_NAME, req_id, [])
        request = yield cursor
        raise gen.Return(request)

    @gen.coroutine
    def load_client_application(self, client_id):
        obj_serv = self.settings['object_service']
        cursor = obj_serv.load(SYSTEM_USER, AUTH_CLIENT_APPLICATION_OBJECT_NAME, client_id, [])
        app = yield cursor
        raise gen.Return(app)

    @gen.coroutine
    def insert_code_status(self, response):
        obj_serv = self.settings['object_service']
        promise = obj_serv.insert(SYSTEM_USER, OAUTH2_CODE_STATUS_OBJECT_NAME, response)
        code_status_id = yield promise
        raise gen.Return(code_status_id)

    @gen.coroutine
    def update_code_status(self, code_status):
        obj_serv = self.settings['object_service']
        promise = obj_serv.save(SYSTEM_USER, OAUTH2_CODE_STATUS_OBJECT_NAME, code_status)
        code_status_id = yield promise
        raise gen.Return(code_status_id)

    @gen.coroutine
    def load_code_status(self, code_status_id):
        obj_serv = self.settings['object_service']
        promise = obj_serv.load(SYSTEM_USER, OAUTH2_CODE_STATUS_OBJECT_NAME, code_status_id, [])
        code_status = yield promise
        raise gen.Return(code_status)

    @gen.coroutine
    def load_user(self, user_id):
        obj_serv = self.settings['object_service']
        cursor = obj_serv.load(SYSTEM_USER, COSMOS_USERS_OBJECT_NAME, user_id, [])
        user = yield cursor
        raise gen.Return(user)

    @gen.coroutine
    def is_trusted_redirect_url(self, full_redirect_url):
        parts = list(urlparse.urlparse(full_redirect_url))
        parts[3] = ''
        parts[4] = ''

        redirect_url = urlparse.urlunparse(parts)

        oauth2_settings = self.settings['oauth2_settings']
        trusted_redirect_urls = oauth2_settings.get("oauth2_trusted_redirect_urls", [])

        raise gen.Return(redirect_url in trusted_redirect_urls)

    def get_private_key_pem(self):
        oauth2_settings = self.settings['oauth2_settings']
        return oauth2_settings.get("oauth2_private_key_pem")



# -------------------------------------- #
# Copyright (C) 2014 Maruf Maniruzzaman  #
# License :: OSI Approved :: MIT License #
# Website: http://cosmosframework.com    #
# Author: Maruf Maniruzzaman             #
# -------------------------------------- #





"""
 Copyright (C) 2016 Maruf Maniruzzaman
 Website: http://cosmosframework.com
 Author: Maruf Maniruzzaman
 License :: OSI Approved :: MIT License
"""

from celery import Celery

def create_app(name, params):
    app = Celery(name, broker=params.get("broker_urls"))
    return app


"""
 Copyright (C) 2014 Maruf Maniruzzaman
 Website: http://cosmosframework.com
 Author: Maruf Maniruzzaman
 License :: OSI Approved :: MIT License
"""

COSMOS_GLOBAL_SETTINGS_OBJECT_NAME = "cosmos.globalsettings"
COSMOS_APPLICATION_OBJECT_NAME = "cosmos.applications"
COSMOS_APPLICATION_ENDPOINT_LIST_OBJECT_NAME = "cosmos.appendpoints"
COSMOS_INTERCEPTOR_OBJECT_NAME = "cosmos.interceptors"
COSMOS_WIDGETS_OBJECT_NAME = "cosmos.widgets"
COSMOS_SOURCE_MODULES_OBJECT_NAME = "cosmos.sourcemodules"
COSMOS_SOURCE_MODULES_TYPE_EMBEDDED = "embedded"
COSMOS_SOURCE_MODULES_TYPE_GRIDFILE = "gridfile"
COSMOS_WORKFLOWS_OBJECT_NAME = "cosmos.workflows"


#Application export / import constants
COSMOS_OBJECT_DATA_FILE_NAME = "object_data.json"
COSMOS_APPLICATION_FILE_NAME = 'application.json'
COSMOS_CURRENT_ARCHIVE_VERSION = "0.01.000.00"

COSMOS_SOURCE_FILE_ROOT_NAME = "sources"